import abc

import requests
import six
from requests import Response
from requests.auth import HTTPBasicAuth
from urlobject import URLObject


@six.add_metaclass(abc.ABCMeta)
class BaseApiClient(object):
    def __init__(self, host=None, *args, **kwargs):
        from shortesttrack_tools.api_client.endpoints.endpoint_urls import EndpointURLs
        self.logger = self._get_logger()
        self.endpoint_urls = EndpointURLs.init(host=host)
        self.host = self.endpoint_urls.HOST
        self.GET = 'GET'
        self.POST = 'POST'
        self.PATCH = 'PATCH'
        self.DELETE = 'DELETE'

    @abc.abstractmethod
    def _get_logger(self):
        """
        Return a logger object

        """

    @abc.abstractmethod
    def _raise_http_error(self, exception: requests.HTTPError, message):
        """
        Raise appropriate exception when HTTP error is occurred

        """

    @abc.abstractmethod
    def _raise_performing_request_error(self, message):
        """
        Raise appropriate exception when Request cannot be performed

        """

    @abc.abstractmethod
    def _get_auth_string(self):
        """
        Return auth user string
        """

    def _construct_response_message(self, response: Response) -> str:
        message = f'Response:'
        message += f' [{response.status_code}]'

        if response.reason:
            message += f' [{response.reason}]'

        if response.content:
            try:
                decoded_content = response.content.decode()
            except Exception as e:
                decoded_content = f' (ERROR: content cannot be decoded: error {e})'
            else:
                decoded_content = ' ' + decoded_content

            decoded_content = decoded_content[:2048]
            message += decoded_content
        else:
            message += ' (ERROR: no content in response)'

        return message

    def _construct_http_error_message(self, response: Response) -> str:
        message = 'HTTPError: '
        message += self._construct_response_message(response)
        return message

    def request(
            self,
            http_method,
            path,
            base: str,
            params=None,
            data=None,
            json=None,
            basic_auth_tuple=None,
            sec_id_for_special_token=None,
            extra_headers=None,
            raw_content=False
    ):
        assert not (basic_auth_tuple and sec_id_for_special_token), 'Only one of these args can be specified'

        if path.startswith('/'):
            path = path[1:]

        retry = False
        auth = None

        if basic_auth_tuple:
            # The basic auth can be used instead of normal Bearer auth
            auth = HTTPBasicAuth(*basic_auth_tuple)
            headers = {}
        else:
            # Normal Bearer auth
            if sec_id_for_special_token:
                assert hasattr(self, 'get_sec_access_token'), 'Method does not exist. Forgot Mixin?'
                # This param is overriding auth with SEC token auth
                headers = {'Authorization': 'Bearer {}'.format(self.get_sec_access_token(sec_id_for_special_token))}
                retry = True
            else:
                # Use user auth
                headers = {'Authorization': self._get_auth_string()}
                retry = False

        url = URLObject(base).add_path(path)
        headers['Content-Type'] = 'application/json'

        if extra_headers:
            headers.update(extra_headers)

        def _request():
            try:
                calling_service_message = f'Calling service {http_method} : {url} : {headers}'
                self.logger.info(calling_service_message)
                resp_ = requests.request(http_method, url,
                                         params=params, data=data, json=json, headers=headers, auth=auth)
            except requests.RequestException as e_:
                self.logger.error('Outer service request error: {}'.format(e_))
                self._raise_performing_request_error('Error on request performing: {}'.format(e_))

            self.logger.info(self._construct_response_message(resp_))
            return resp_

        resp = _request()
        try:
            resp.raise_for_status()
        except requests.HTTPError as e1:
            if retry and resp.status_code in [401, 403]:
                self._retry_occurred()
                # Probably sec_access_token is expired, let's get a new one
                headers = {'Authorization': 'Bearer {}'.format(self.get_sec_access_token(sec_id_for_special_token))}
                # Retry request
                resp = _request()
                try:
                    resp.raise_for_status()
                except requests.HTTPError as e2:
                    # This is not access_token expiration, raise exception
                    http_error_message = self._construct_http_error_message(resp)
                    self._raise_http_error(e2, http_error_message)
            else:
                http_error_message = self._construct_http_error_message(resp)
                self._raise_http_error(e1, http_error_message)

        except Exception as e:
            # This error should never be happened
            self.logger.error('Unexpected error (it can be error in base_api_client module): {}'.format(e))
            raise

        if not raw_content:
            return resp.json()

        return resp.content

    def _retry_occurred(self):
        """
        This method is overridden for test purposes. By default it's do nothing and don't need to be overridden.
        """



from urlobject import URLObject
import os

from shortesttrack_tools.unique import Unique


class EndpointURLs(Unique):
    HOST = None

    UAA_SERVICE_ENDPOINT = None
    OAUTH_SERVICE_ENDPOINT = None
    JWTKEY_ENDPOINT = None
    METADATA_SERVICE_ENDPOINT = None
    EXEC_API_SERVICE_ENDPOINT = None
    EXEC_SCHEDULER_SERVICE_ENDPOINT = None
    DATA_SERVICE_ENDPOINT = None

    @classmethod
    def _do_init(cls, host=None):
        if host is None:
            cls.HOST = HOST = URLObject(os.environ.get('HOST', 'https://shortesttrack.com'))
        else:
            cls.HOST = HOST = URLObject(host)

        cls.UAA_SERVICE_ENDPOINT = HOST.add_path('api/uaa')
        cls.OAUTH_SERVICE_ENDPOINT = HOST.add_path('oauth')
        cls.JWTKEY_ENDPOINT = HOST.add_path('oauth_jwtkey')
        cls.METADATA_SERVICE_ENDPOINT = HOST.add_path('api/metadata')
        cls.EXEC_API_SERVICE_ENDPOINT = HOST.add_path('api/execution-metadata/v2')
        cls.EXEC_SCHEDULER_SERVICE_ENDPOINT = HOST.add_path('api/exec-scheduling')
        cls.DATA_SERVICE_ENDPOINT = HOST.add_path('api/data')



from .base_api_client import BaseApiClient
from .sec_refresh_token_mixin import SecRefreshTokenMixin
from .sec_access_token_mixin import SecAccessTokenMixin
from .public_key_mixin import PublicKeyMixin



# noinspection PyMethodOverriding,PyArgumentList,PyPep8Naming
class classproperty(property):
    def __init__(self, fget, *arg, **kw):
        super(classproperty, self).__init__(fget, *arg, **kw)
        self.__doc__ = fget.__doc__

    def __get__(desc, self, cls):
        return desc.fget(cls)



from collections import defaultdict

from ...api_client import BaseApiClient, SecRefreshTokenMixin, SecAccessTokenMixin


class MockLogger(object):
    def __getattr__(self, item):
        def _mock(*args, **kwargs):
            pass

        return _mock


logger = MockLogger()


class MockApiClient(BaseApiClient):
    class HttpError(Exception):
        pass

    class RequestError(Exception):
        pass

    def _raise_performing_request_error(self, *args, **kwargs):
        raise self.RequestError(*args)

    def _raise_http_error(self, *args, **kwargs):
        raise self.HttpError(*args)

    def _get_auth_string(self):
        return 'Bearer foo'

    def _get_logger(self):
        return logger


class MockRefreshTokenApiClient(SecRefreshTokenMixin, MockApiClient):
    def __init__(self,  *args, **kwargs):
        super(MockRefreshTokenApiClient, self).__init__(*args, **kwargs)
        self.mock_store = defaultdict(lambda: defaultdict(dict))

    def _set_cache_sec_refresh_token(self, configuration_id, refresh_token):
        self.mock_store[configuration_id]['refresh_token'] = refresh_token

    def _get_cache_sec_refresh_token(self, configuration_id):
        return self.mock_store[configuration_id]['refresh_token']


class MockAccessTokenApiClient(SecAccessTokenMixin, MockApiClient):
    def __init__(self, *args, **kwargs):
        super(MockAccessTokenApiClient, self).__init__(*args, **kwargs)
        self.mock_store = defaultdict(lambda: defaultdict(dict))

    def _get_cache_public_key(self):
        return self.mock_store['public_key'][0]

    def _set_cache_public_key(self, public_key):
        self.mock_store['public_key'][0] = public_key

    def _set_cache_sec_access_token_expiration(self, configuration_id, expiration):
        self.mock_store[configuration_id]['expiration'] = expiration

    @property
    def _basic_auth_tuple(self):
        return ('foo', 'bar',)

    def _get_cache_sec_access_token_expiration(self, configuration_id):
        return self.mock_store[configuration_id]['expiration']

    def _set_cache_sec_access_token(self, configuration_id, access_token):
        self.mock_store[configuration_id]['access_token'] = access_token

    def get_sec_refresh_token(self, configuration_id):
        return 'REFRESH_TOKEN'

    def _get_cache_sec_access_token(self, configuration_id):
        return self.mock_store[configuration_id]['access_token']



from shortesttrack_tools.api_client import BaseApiClient


class BaseEndpoint(object):
    def __init__(self, api_client: BaseApiClient, base, script_execution_configuration_id, *args, **kwargs):
        self._api_client = api_client
        self.endpoint_urls = api_client.endpoint_urls
        self._base = base
        self._script_execution_configuration_id = script_execution_configuration_id
        self.GET = 'GET'
        self.POST = 'POST'
        self.PATCH = 'PATCH'
        self.DELETE = 'DELETE'

    def request(
            self,
            http_method,
            path,
            params=None,
            data=None,
            json=None,
            extra_headers=None,
            raw_content=False
    ):
        kwargs = {
            'http_method': http_method,
            'path': path,
            'params': params,
            'data': data,
            'json': json,
            'base': self._base,
            'basic_auth_tuple': None,
            'sec_id_for_special_token': self._script_execution_configuration_id,
            'extra_headers': extra_headers,
            'raw_content': raw_content
        }

        return self._api_client.request(**kwargs)



import json

import jwt
from jwt import InvalidTokenError


class JWTUtil(object):
    class InvalidTokenError(Exception):
        pass

    def __init__(self, jwt_token):
        self.jwt_token = jwt_token

    def decode(self, public_key, verify_exp=False):
        try:
            decoded = jwt.decode(self.jwt_token, public_key,
                                 algorithms='RS256', options={'verify_exp': verify_exp}, audience=['web_app'])
        except InvalidTokenError as e:
            raise self.InvalidTokenError(str(e))
        return decoded

    def authenticate(self, public_key):
        return self.decode(public_key, True)

    def extract_info_no_verify(self):
        return jwt.decode(self.jwt_token, verify=False)

    def extract_expiration_string_no_verify(self):
        decoded = self.extract_info_no_verify()
        return decoded['exp']

    def extract_expiration_string(self, public_key):
        decoded = self.decode(public_key)
        return decoded['exp']

    @staticmethod
    def encode(obj_or_filename, private_key_or_filename, obj_from_file=False, private_key_from_file=False):
        if obj_from_file:
            with open(obj_or_filename, 'r') as f:
                obj = json.load(f)
        else:
            obj = obj_or_filename

        if private_key_from_file:
            with open(private_key_or_filename, 'r') as f:
                private_key = f.read()
        else:
            private_key = private_key_or_filename

        return jwt.encode(obj, private_key, 'RS256')
